// zad5.cpp
// g++ -o prog zad5.cpp -pthread

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <semaphore.h>
#include <iostream>

using namespace std;

// objekat tipa Buffer ce biti deljen medju nitima
class Buffer{
private:
	int* buf;
	int head, tail, cap;
	sem_t spaceAvailable, itemAvailable;
	sem_t mutexPut, mutexGet;
public:
	Buffer(int cap){
		buf = new int[cap];
		this->cap = cap;
		head = tail = 0;
		sem_init(&spaceAvailable, 0, cap);
		sem_init(&itemAvailable, 0, 0);
		sem_init(&mutexPut, 0, 1);
		sem_init(&mutexGet, 0, 1);
	}
	~Buffer(){
		delete buf;
		sem_destroy(&spaceAvailable);
		sem_destroy(&itemAvailable);
		sem_destroy(&mutexGet);
		sem_destroy(&mutexPut);
	}
	int put(int item){
		sem_wait(&spaceAvailable);
		sem_wait(&mutexPut);		
		buf[head] = item;
		head = (head + 1) % cap;
		sem_post(&mutexPut);		
		sem_post(&itemAvailable);
		return item;
	}
	int get(){
		sem_wait(&itemAvailable);
		sem_wait(&mutexGet);		
		int ret = buf[tail];
		tail = (tail + 1) % cap;
		sem_post(&mutexGet);
		sem_post(&spaceAvailable);
		return ret;
	}
};

// nit 1 - proizvodjac
void* putThread(void* args){
	Buffer* b = (Buffer*)args;
	int iters = 10;
	srand(time(0));

	while(iters-- > 0){
		usleep( rand() % 1000 * 1000 );
		//printf("Put: %d\n", b->put(rand()%10));
		cout << "Put: " << b->put(rand()%10) << endl;		
		//fflush(stdout);
		cout.flush();
		
	}
}

// nit 2 - potrosac
void* getThread(void* args){
	Buffer* b = (Buffer*)args;
	int iters = 10;
	
	while(iters-- > 0){
		//printf("Get: %d\n", b->get());
	  cout << "Get: " << b->get() << endl;
	}
}

int main(){
	// deljeni objekat bafera sa kapacitetom 2
	Buffer b(2);

	pthread_t threads[2];
	pthread_create(&threads[0], 0, putThread, &b);
	pthread_create(&threads[1], 0, getThread, &b);

	pthread_join(threads[0], 0);
	pthread_join(threads[1], 0);

	printf("Happy ending!\n");

	return 0;
}
