// zad5.cpp - drugi nacin
// g++ -o prog zad5.cpp -pthread

#include <iostream>
#include <pthread.h>
#include <semaphore.h>
#include <cstdlib>

#define PRODS_CNT 10
#define CONS_CNT 1

using namespace std;

class Buffer{
public:
	Buffer(int cap = 10);
	~Buffer();
	void put(int item);
	int get();
private:
	int head, tail, cap;
	int* buf;
	sem_t itemAvail, slotsAvail, mutexPut, mutexGet;
};

Buffer::Buffer(int cap){
	buf = new int[this->cap = cap];
	head = tail = 0;
	sem_init(&itemAvail, 0, 0);
	sem_init(&slotsAvail, 0, cap);
	sem_init(&mutexPut, 0, 1);
	sem_init(&mutexGet, 0, 1);
}

Buffer::~Buffer(){
	sem_destroy(&itemAvail);
	sem_destroy(&slotsAvail);
	sem_destroy(&mutexPut);
	sem_destroy(&mutexGet);
	delete []buf;
}

void Buffer::put(int item){
	sem_wait(&slotsAvail); // slotsAvail.wait -- cap
	sem_wait(&mutexPut); // mutexPut.wait -- 1
	buf[head] = item;
	head = (head + 1) % cap;
	sem_post(&mutexPut); // mutexPut.signal
	sem_post(&itemAvail); // itemAvail.signal -- 0
}

int Buffer::get(){
	sem_wait(&itemAvail); // itemAvail.wait
	sem_wait(&mutexGet); // mutexGet.wait
	int item = buf[tail];
	tail = (tail + 1) % cap;
	sem_post(&mutexGet); // mutexGet.signal
	sem_post(&slotsAvail);  // slotsAvail.signal
	return item;
}

sem_t glblMutex; // sluzi pri pozivanju cout << 

void* produce(void* params){
	void** data_array = (void**)params;
	Buffer* buf = (Buffer*)data_array[0];
	int* item = (int*)data_array[1];
	int i = 3;	
	while(i-- > 0){
		buf->put(*item);
		sem_wait(&glblMutex);
			cout << "put: "<< *item << endl;
		sem_post(&glblMutex);		
	}
	delete (int*)(data_array[1]);
	delete []data_array;
	return 0;
}

void* consume(void* params){
	Buffer* buf = (Buffer*)params;
	int i = 3 * PRODS_CNT;
	while(i-- > 0){
		int item = buf->get();
		sem_wait(&glblMutex);
			cout << "get: " << item << endl;
		sem_post(&glblMutex);	
	}
	return 0;
}

int main(int argc, char* argv[]){
	if(argc < 2){
		cout << "Usage: ./prog cap" << endl;
		return 1;	
	}
	int cap = atoi(argv[1]);
	Buffer buf(cap);	

	sem_init(&glblMutex, 0, 1);

	pthread_t thProds[PRODS_CNT];
	pthread_t thCons[CONS_CNT];

	for(int i = 0; i < PRODS_CNT; i++){
		void* *data = new void*[2];
		data[0] = (void*)(&buf);
		data[1] = (void*)(new int(i));
		pthread_create(&thProds[i], 0, produce, (void*)data);
	}

	for(int i = 0; i < CONS_CNT; i++){
		pthread_create(&thCons[i], 0, consume, (void*)(&buf));
	}


	for(int i = 0; i < PRODS_CNT + CONS_CNT; i++){
		if(i < PRODS_CNT){
			pthread_join(thProds[i], 0);		
		}
		else {
			pthread_join(thCons[i - PRODS_CNT], 0);			
		}
	}
	
	sem_destroy(&glblMutex);
	return 0;
	
}
